use crate::vec_util::IVec2Helper;
use crate::{Image8, Tileset};
use glam::IVec2;
use rand::distributions::Standard;
use rand::prelude::Distribution;
use rand::Rng;
use std::rc::Rc;

//odd-q vertical layout https://www.redblobgames.com/grids/hexagons
pub struct HexMap {
    size: IVec2,
    data: Vec<i32>,
    tileset: Rc<Tileset>,
    pix_tile_off: IVec2,
}

pub enum Direction {
    North,
    NorthEast,
    SouthEast,
    South,
    SouthWest,
    NorthWest,
}

impl HexMap {
    //pub static
    //pix_tile_off: x offset per tile, up/down alternating y offset on x axis
    pub fn new(size: IVec2, tileset: Rc<Tileset>, pix_tile_off: IVec2) -> Self {
        HexMap {
            size,
            data: vec![0; size.size()],
            tileset,
            pix_tile_off,
        }
    }

    pub fn get_neighbour(coord: IVec2, dir: Direction) -> IVec2 {
        let yoff = coord.x % 2;
        (match dir {
            Direction::North => IVec2 { x: 0, y: -1 },
            Direction::NorthEast => IVec2 { x: 1, y: -1 + yoff },
            Direction::SouthEast => IVec2 { x: 1, y: yoff },
            Direction::South => IVec2 { x: 0, y: 1 },
            Direction::SouthWest => IVec2 { x: -1, y: yoff },
            Direction::NorthWest => IVec2 {
                x: -1,
                y: -1 + yoff,
            },
        }) + coord
    }

    //pub
    pub fn get(&self, coord: IVec2) -> i32 {
        if !self.is_valid_coord(coord) {
            -1
        } else {
            self.data[self.coord_to_idx(coord)]
        }
    }

    pub fn pixel_to_coord(&self, pixel: IVec2) -> IVec2 {
        let tilesize = self.tileset.tile_size();
        let xrepeat = pixel.x % self.pix_tile_off.x;
        let mut x = pixel.x / self.pix_tile_off.x;
        let py = if x % 2 == 0 {
            pixel.y
        } else {
            pixel.y - self.pix_tile_off.y
        };
        let yrepeat = py % tilesize.y;
        let mut y = py / tilesize.y;
        let outside = self.tileset.get(0).get_pixel(IVec2 {
            x: xrepeat,
            y: yrepeat,
        }) == 0;
        if outside {
            x -= 1;
            if yrepeat > tilesize.y / 2 {
                y += 1;
            }
            y -= x % 2;
        }
        IVec2 { x, y }
    }

    fn coord_to_idx(&self, coord: IVec2) -> usize {
        (coord.x + coord.y * self.size.x) as usize
    }

    pub fn coord_to_pixel(&self, coord: IVec2) -> IVec2 {
        IVec2 {
            x: coord.x * self.pix_tile_off.x,
            y: coord.y * self.tileset.tile_size().y + (coord.x % 2) * self.pix_tile_off.y,
        }
    }

    pub fn coord_to_pixel_center(&self, coord: IVec2) -> IVec2 {
        let tilesize = self.tileset.tile_size();
        self.coord_to_pixel(IVec2 {
            x: coord.x + tilesize.x / 2,
            y: coord.y + tilesize.y / 2,
        })
    }

    pub fn is_valid_coord(&self, coord: IVec2) -> bool {
        coord.x >= 0 && coord.y >= 0 && coord.x < self.size.x && coord.y < self.size.y
    }

    pub fn set(&mut self, coord: IVec2, val: i32) {
        let idx = self.coord_to_idx(coord);
        self.data[idx] = val;
    }

    pub fn size(&self) -> IVec2 {
        self.size
    }

    pub fn idx_size(&self) -> usize {
        self.size.size()
    }

    //pubcrate
    pub(crate) fn draw_tile_to_image(&self, target: &mut Image8, coord: IVec2, offset: IVec2) {
        target.draw_image(
            self.coord_to_pixel(coord) + offset,
            self.tileset.get(self.data[self.coord_to_idx(coord)]),
        )
    }
}

impl Distribution<Direction> for Standard {
    fn sample<R: Rng + ?Sized>(&self, rng: &mut R) -> Direction {
        match rng.gen_range(0..6) {
            0 => Direction::North,
            1 => Direction::NorthEast,
            2 => Direction::SouthEast,
            3 => Direction::South,
            4 => Direction::SouthWest,
            _ => Direction::NorthWest,
        }
    }
}
